import time
from itertools import chain
from typing import List, Optional, Tuple, Union

import navis
import networkx as nx
import numpy as np
import plotly.graph_objs as go
from matplotlib import pyplot as plt
from plotly.offline import iplot, plot

from neuroboom.utils import calc_cable, check_valid_neuron_input

# from logging import
# This script contains functions for plotting dendrograms, static and interactive


def create_graph_structure(
    neuron: Union[navis.TreeNeuron, navis.NeuronList],
    xyz_locs: bool = True,
    returned_object: str = "graph",
    prog: str = "dot",
    verbose: bool = False
):

    """
    Takes a navis neuron and creates a graph layout

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """

    neuron = check_valid_neuron_input(neuron)

    valid_objects = ["graph", "positions", "graph_and_positions"]

    assert isinstance(
        returned_object, str
    ), f"You need to pass a string for the returned object. You have passed: {returned_object}"

    assert (
        returned_object in valid_objects
    ), f"You need to pass a valid object to return. These include: {valid_objects}"

    valid_progs = ["fdp", "dot", "neato"]

    assert (
        prog in valid_progs
    ), f"Invalid program parameter. You need to pass one of {valid_progs}"

    if verbose:
        print("Creating Graph Structure...")

    if 'parent_dist' not in neuron.nodes:

        if verbose:
            print("Calculating cable lengths...")
        neuron = calc_cable(neuron, return_skdata=True)

    g = nx.DiGraph()
    g.add_nodes_from(neuron.nodes.node_id)

    for e in neuron.nodes[['node_id','parent_id','parent_dist']].values:
        # skip root node
        if e[1] == -1:
            continue
        g.add_edge(int(e[0]), int(e[1]), len=e[2])
        #g.add_edge(int(e[0]), int(e[1]), normal=e[2])

        if xyz_locs:

            vals = neuron.nodes[['x','y','z']].values
            pos_dict = dict(zip(neuron.nodes.node_id, vals))

            nx.set_node_attributes(g, pos_dict, 'pos')

    if returned_object == "graph":
        if verbose:
            print("Returning graph only")
        return g

    elif returned_object == "positions":
        if verbose:
            print("Calculating layout...")
        pos = nx.nx_agraph.graphviz_layout(g, prog=prog)
        if verbose:
            print("Returning positions only")
        return pos

    elif returned_object == "graph_and_positions":
        if verbose:
            print("Calculating layout...")
        pos = nx.nx_agraph.graphviz_layout(g, prog=prog)
        if verbose:
            print("Returning graph and positions")
        return (g, pos)


# dendrogram


def plot_dendrogram(
    x: Union[navis.TreeNeuron, navis.NeuronList],
    heal_neuron: bool = False,
    downsample_neuron: float = 0.0,
    plot_connectors: bool = True,
    connector_confidence: Tuple[float, float] = (0.0, 0.0),
    highlight_nodes: Optional = None,
    highlight_connectors: Optional = None,
    fragment: bool = False,
    presyn_color: List[List[float]] = [[0.9, 0.0, 0.0]],
    postsyn_color: List[List[float]] = [[0.0, 0.0, 0.9]],
    highlight_connector_color: List[List[float]] = [[0.0, 0.9, 0.0]],
    highlight_connector_size: int = 20,
    presyn_size: float = 0.1,
    postsyn_size: float = 0.1,
    prog: str = "dot",
):

    """
    This function creates a 2-dimensional dendrogram, a 'flattened' version of a neuron.
    Dendrograms can be used to visualise the locations of specific partner synapses.

    Parameters
    ----------
    x :                navis.TreeNeuron

                        A single navis tree neuron object

    heal_neuron :      bool

                        Whether you want to heal the neuron or not. N.B. Navis neurons
                        should be healed on import, i.e. navis.fetch_skeletons(bodyid, heal = True)
                        see navis.fetch_skeletons and navis.heal_fragmented_neuron for more details

    downsample_neuron: float

                        A float specifying the downsampling factor used by navis.downsample_neuron()
                        If 0.0, then no downsampling will occur. If float('inf') then this will reduce
                        the neuron to branch and end points.

                        It is recommended to downsample very large neurons when testing out this code
                        for the first time.

    plot_connectors:   bool

                        Whether to plot presynapses and postsynapses on the dendrogram or not.

    connector_confidence: tuple

                        The confidence value used to threshold the synapses.
                        The first value (connector_confidence[0]) will be used to threshold presynapses
                        The second value (connector_confidence[1]) will be used to threshold postsynapses

    highlight_connectors: optional | np.array | dict

                        If a numpy array, then this should be an array of the treenodes
                        connected to the connectors that you want to highlight.
                        The single color and size will be specified in
                        highlight_connector_color and highlight_connector_size

                        If a dictionary, then the key values should be
                        treenode ids of the connectors you want to highlight
                        and their values should be the color you want to colour them.

                        Passing dictionaries to this parameter allow for synapses to be coloured differently

    fragment:           bool

                        Whether the neuron object you are passing
                        is a fragment or not (i.e. does it have a soma or not)

    presyn_color:       list

                        A list containing the rgb values that you want to colour the presynapses.
                        All presynapses will be coloured this color

    postsyn_color:      list

                        A list containing the rgb values that you want to colour the postsynapses.
                        All postsynapses will be coloured this color

    highlight_connector_color: list

                        A list containing the rgb values that you want to color your special synapses

    highlight_connector_size: int

                        The size of the synapses you want to highlight on the dendrogram

    presyn_size:         int

                        The size of all presynapses on the dendrogram

    postsyn_size:        int

                        The size of all postsynapses on the dendrogrm

    prog :             str

                       The layout type used by navis.nx_agraph.graphviz_layout()
                       Valid programs include [dot, neato or fdp].
                       The dot program provides a hierarchical layout, this is the fastest program
                       The neato program creates edges between nodes proportional to their real length.
                       The neato program takes the longest amount of time, can be ~2hrs for a single neuron!


    Returns
    -------
    fig: a figure of containing the dendrogram


    Example
    --------
    from neuroboom.utils import create_graph_structure
    from neuroboom.dendrogram import plot_dendrogram
    import navis.interfaces.neuprint as nvneu
    from matplotlib import pyplot as plt

    test_neuron = nvneu.fetch_skeletons(722817260)

    plt.clf()
    fig, ax = plt.subplots(figsize = (20,20))
    plot_dendrogram(test_neuron, prog = 'dot')
    plt.show()

    """

    x = check_valid_neuron_input(x)

    assert isinstance(
        connector_confidence, tuple
    ), f"Need to pass a tuple for confidence values. You have passed a {type(connector_confidence)}"

    assert (
        len(connector_confidence) == 2
    ), """
    Need to pass a tuple containing two values for confidence. \n
    The first value is the confidence threshold for presynapses. \n
    The second value is the confidence threshold for postsynapses. """

    valid_progs = ["fdp", "dot", "neato"]

    assert (
        prog in valid_progs
    ), f"Invalid program parameter. You need to pass one of {valid_progs}"

    start = time.time()

    if heal_neuron:
        print("Healing Neuron...")
        navis.heal_fragmented_neuron(x, inplace=True)

    if any(connector_confidence) > 0.0:
        print(
            """Thresholding synapses: only considering
        presynapses above {} confidence and postsynapses above {}""".format(
                connector_confidence[0], connector_confidence[1]
            )
        )
        presyn_included = x.connectors[x.connectors.type == "pre"][
            x.connectors.confidence > connector_confidence[0]
        ].connector_id.tolist()
        postsyn_included = x.connectors[x.connectors.type == "post"][
            x.connectors.confidence > connector_confidence[1]
        ].connector_id.tolist()
        connectors_included = list(
            chain.from_iterable([presyn_included, postsyn_included])
        )
        x.connectors = x.connectors[x.connectors.connector_id.isin(connectors_included)]

    if downsample_neuron > 0:
        print("Downsampling neuron, factor = {}".format(downsample_neuron))
        x = navis.downsample_neuron(
            x,
            downsampling_factor=downsample_neuron,
            preserve_nodes=x.connectors.node_id.unique().tolist(),
        )

    if "parent_dist" not in x.nodes:
        print("Calculating cable length...")
        x = calc_cable(x, return_skdata=True)

    g, pos = create_graph_structure(x, returned_object="graph_and_positions", prog=prog)

    # Plotting tree with the above layout
    print("Plotting Tree...")
    nx.draw(g, pos, node_size=0, arrows=False, width=0.25)

    # Whether to add soma or not
    if not fragment:

        if type(x.soma) == np.ndarray:

            print("Plotting soma")
            soma = x.soma[0]
            plt.scatter([pos[soma][0]], [pos[soma][1]], s = 80, c = [[0, 0, 0]], zorder = 1)

        else:

            print("Plotting soma")
            plt.scatter([pos[x.soma][0]], [pos[x.soma][1]], s=80, c=[[0, 0, 0]], zorder=1)

    if plot_connectors:
        print("Plotting connectors...")
        plt.scatter(
            [
                pos[tn][0]
                for tn in x.connectors[x.connectors.type == "pre"].node_id.values
            ],
            [
                pos[tn][1]
                for tn in x.connectors[x.connectors.type == "pre"].node_id.values
            ],
            c=presyn_color,
            zorder=2,
            s=presyn_size,
            linewidths=1,
        )

        plt.scatter(
            [
                pos[tn][0]
                for tn in x.connectors[x.connectors.type == "post"].node_id.values
            ],
            [
                pos[tn][1]
                for tn in x.connectors[x.connectors.type == "post"].node_id.values
            ],
            c=postsyn_color,
            zorder=2,
            s=postsyn_size,
            linewidths=1,
        )

    if highlight_nodes is not None:

        if isinstance(highlight_nodes, dict):

            hl_tn_coords = np.array(
                [
                    pos[tn] for tn in highlight_nodes.keys()

                ]
            )

            tn_col = [highlight_nodes[i] for i in highlight_nodes.keys()]

            plt.scatter(
                hl_tn_coords[:, 0],
                hl_tn_coords[:, 1],
                s = 10,
                c = tn_col,
                zorder = 3
            )

    if highlight_connectors is not None:

        if isinstance(highlight_connectors, (list, np.ndarray)):
            hl_cn_coords = np.array(
                [
                    pos[tn]
                    for tn in x.connectors[
                        x.connectors.connector_id.isin(highlight_connectors)
                    ].node_id
                ]
            )
            plt.scatter(
                hl_cn_coords[:, 0],
                hl_cn_coords[:, 1],
                s=highlight_connector_size,
                c=highlight_connector_color,
                zorder=3,
                linewidths=1,
            )

        elif isinstance(highlight_connectors, dict):
            for cn in highlight_connectors:
                if cn in highlight_connectors:
                    if cn is None:
                        continue
                    if cn not in x.connectors.connector_id.values:
                        print(
                            "Connector {} is not present in the neuron / neuron fragment".format(
                                cn
                            )
                        )
                    hl_cn_coords = np.array(
                        [
                            pos[tn]
                            for tn in x.connectors[
                                x.connectors.connector_id == cn
                            ].node_id
                        ]
                    )
                    plt.scatter(
                        hl_cn_coords[:, 0],
                        hl_cn_coords[:, 1],
                        s=highlight_connector_size,
                        color=highlight_connectors[cn],
                        zorder=3,
                    )

        else:

            raise TypeError(
                "Unable to highlight connectors from data of type {}".format(
                    type(highlight_connectors)
                )
            )

    print("Completed in %is" % int(time.time() - start))


# Plot an interactive dendrogram
def interactive_dendrogram(
    z: Union[navis.TreeNeuron, navis.NeuronList],
    heal_neuron: bool = True,
    plot_nodes: bool = True,
    plot_connectors: bool = True,
    highlight_connectors: Optional = None,
    in_volume: Optional = None,
    prog: str = "dot",
    inscreen: bool = True,
    filename: Optional = None,
):

    """
    Takes a navis neuron and returns an interactive 2D dendrogram.
    In this dendrogram, nodes or connector locations can be highlighted


    Parameters
    ----------
    x:                    A navis neuron object

    heal_neuron: bool

                    Whether you want to heal the neuron or not. N.B. Navis neurons
                    should be healed on import, i.e. navis.fetch_skeletons(bodyid, heal = True)
                    see navis.fetch_skeletons and navis.heal_fragmented_neuron for more details

    plot_nodes:           bool

                    Whether treenodes should be plotted

    plot_connectors:      bool

                    Whether connectors should be plotted

    highlight_connectors: dict

                    A dictionary containing the treenodes of
                    the connectors you want to highlight as keys
                    and the colours you want to colour them as values.
                    This allows for multiple colours to be plotted.

                    N.B. Plotly colours are in the range of 0 - 255
                    whereas matplotlib colours are between 0-1. For the
                    interactive dendrogram colours need to be in the
                    plotly range, whereas in the static dendrogram
                    the colours need to be in the matplotlib range.

    in_volume:            navis.Volume object

                    A navis.Volume object corresponding to an ROI in the brain.
                    This will then highlight the nodes of
                    the neuron which are in that volume


    prog:                 str

                    The layout type used by navis.nx_agraph.graphviz_layout()
                    Valid programs include [dot, neato or fdp].
                    The dot program provides a hierarchical layout, this is the fastest program
                    The neato program creates edges between nodes proportional to their real length.
                    The neato program takes the longest amount of time, can be ~2hrs for a single neuron!

    inscreen:             bool

                    Whether to plot the graph inscreen (juptyer notebooks) or to plot it as a
                    separate HTML file that can be saved to file, opened in the browser and opened any time

    filename:             str

                    The filename of your interactive dendrogram html file. This parameter is only appropriate
                    when inscreen = False.



    Returns
    -------
    plotly.fig object containing the dendrogram - this can be either inscreen or as a separate html file
    with the filename specified by the filename parameter


    Examples
    --------
    from neuroboom.utils import create_graph_structure
    from neuroboom.dendrogram import interactive_dendrogram
    import navis.interfaces.neuprint as nvneu
    from matplotlib import pyplot as plt


    test_neuron = nvneu.fetch_skeletons(722817260)


    interactive_dendrogram(test_neuron, prog = 'dot', inscreen = True)


    """

    z = check_valid_neuron_input(z)

    if heal_neuron:
        z = navis.heal_fragmented_neuron(z)

    valid_progs = ["neato", "dot"]
    if prog not in valid_progs:
        raise ValueError("Unknown program parameter!")

    # save start time

    start = time.time()

    # Necessary for neato layouts for preservation of segment lengths

    if "parent_dist" not in z.nodes:
        z = calc_cable(z, return_skdata=True)

    # Generation of networkx diagram

    g, pos = create_graph_structure(z, returned_object="graph_and_positions", prog=prog)

    print("Now creating plotly graph...")

    # Convering networkx nodes for plotly
    # NODES

    x = []
    y = []
    node_info = []

    if plot_nodes:

        for n in g.nodes():
            x_, y_ = pos[n]

            x.append(x_)
            y.append(y_)
            node_info.append("Node ID: {}".format(n))

        node_trace = go.Scatter(
            x=x,
            y=y,
            mode="markers",
            text=node_info,
            hoverinfo="text",
            marker=go.scatter.Marker(showscale=False),
        )
    else:

        node_trace = go.Scatter()

    # EDGES

    xe = []
    ye = []

    for e in g.edges():

        x0, y0 = pos[e[0]]
        x1, y1 = pos[e[1]]

        xe += [x0, x1, None]
        ye += [y0, y1, None]

    edge_trace = go.Scatter(
        x=xe,
        y=ye,
        line=go.scatter.Line(width=1.0, color="#000"),
        hoverinfo="none",
        mode="lines",
    )

    # SOMA

    xs = []
    ys = []

    for n in g.nodes():
        if n != z.soma:
            continue
        elif n == z.soma:

            x__, y__ = pos[n]
            xs.append(x__)
            ys.append(y__)

        else:

            break

    soma_trace = go.Scatter(
        x=xs,
        y=ys,
        mode="markers",
        hoverinfo="text",
        marker=dict(size=20, color="rgb(0,0,0)"),
        text="Soma, node:{}".format(z.soma),
    )

    # CONNECTORS:
    # RELATION  = 0 ARE PRESYNAPSES, RELATION = 1 ARE POSTSYNAPSES

    if plot_connectors is False:

        presynapse_connector_trace = go.Scatter()

        postsynapse_connector_trace = go.Scatter()

    elif plot_connectors is True:

        presynapse_connector_list = list(
            z.connectors[z.connectors.type == "pre"].node_id.values
        )

        x_pre = []
        y_pre = []
        presynapse_connector_info = []

        for node in g.nodes():
            for tn in presynapse_connector_list:

                if node == tn:

                    x, y = pos[node]

                    x_pre.append(x)
                    y_pre.append(y)
                    presynapse_connector_info.append(
                        "Presynapse, connector_id: {}".format(tn)
                    )

        presynapse_connector_trace = go.Scatter(
            x=x_pre,
            y=y_pre,
            text=presynapse_connector_info,
            mode="markers",
            hoverinfo="text",
            marker=dict(size=10, color="rgb(0,255,0)"),
        )

        postsynapse_connectors_list = list(
            z.connectors[z.connectors.type == "post"].node_id.values
        )

        x_post = []
        y_post = []
        postsynapse_connector_info = []

        for node in g.nodes():
            for tn in postsynapse_connectors_list:

                if node == tn:

                    x, y = pos[node]

                    x_post.append(x)
                    y_post.append(y)

                    postsynapse_connector_info.append(
                        "Postsynapse, connector id: {}".format(tn)
                    )

        postsynapse_connector_trace = go.Scatter(
            x=x_post,
            y=y_post,
            text=postsynapse_connector_info,
            mode="markers",
            hoverinfo="text",
            marker=dict(size=10, color="rgb(0,0,255)"),
        )

    if highlight_connectors is None:

        HC_trace = go.Scatter()

    elif isinstance(highlight_connectors, dict):

        HC_nodes = []
        HC_color = []

        for i in list(highlight_connectors.keys()):

            HC_nodes.append(
                z.connectors[z.connectors.connector_id == i].node_id.values[0]
            )
            HC_color.append(
                "rgb({}, {}, {})".format(
                    int(highlight_connectors[i][0]),
                    int(highlight_connectors[i][1]),
                    int(highlight_connectors[i][2]),
                )
            )

            HC_x = []
            HC_y = []

            HC_info = []

            for node in g.nodes():

                for tn in HC_nodes:

                    if node == tn:

                        x, y = pos[node]

                        HC_x.append(x)
                        HC_y.append(y)

                        HC_info.append(
                            "Connector of Interest, connector_id: {}, treenode_id: {}".format(
                                z.connectors[
                                    z.connectors.node_id == node
                                ].connector_id.values[0],
                                node,
                            )
                        )

            HC_trace = go.Scatter(
                x=HC_x,
                y=HC_y,
                text=HC_info,
                mode="markers",
                hoverinfo="text",
                marker=dict(size=12.5, color=HC_color),
            )

    # Highlight the nodes that are in a particular volume

    if in_volume is None:

        in_volume_trace = go.Scatter()

    elif in_volume is not None:

        # Volume of interest

        res = navis.in_volume(z.nodes, volume=in_volume, mode="IN")
        z.nodes["IN_VOLUME"] = res

        x_VOI = []
        y_VOI = []

        VOI_info = []

        for nodes in g.nodes():

            for tn in list(z.nodes[z.nodes.IN_VOLUME == True].node_id.values):

                if nodes == tn:

                    x, y = pos[tn]

                    x_VOI.append(x)
                    y_VOI.append(y)

                    VOI_info.append("Treenode {} is in {} volume".format(tn, in_volume))

        in_volume_trace = go.Scatter(
            x=x_VOI,
            y=y_VOI,
            text=VOI_info,
            mode="markers",
            hoverinfo="text",
            marker=dict(size=5, color="rgb(35,119,0)"),
        )

    print("Creating Plotly Graph")

    fig = go.Figure(
        data=[
            edge_trace,
            node_trace,
            soma_trace,
            presynapse_connector_trace,
            postsynapse_connector_trace,
            HC_trace,
            in_volume_trace,
        ],
        layout=go.Layout(
            title="Plotly graph of {} with {} layout".format(z.name, prog),
            titlefont=dict(size=16),
            showlegend=False,
            hovermode="closest",
            margin=dict(b=20, l=50, r=5, t=40),
            annotations=[
                dict(showarrow=False, xref="paper", yref="paper", x=0.005, y=-0.002)
            ],
            xaxis=go.layout.XAxis(showgrid=False, zeroline=False, showticklabels=False),
            yaxis=go.layout.YAxis(showgrid=False, zeroline=False, showticklabels=False),
        ),
    )

    if inscreen is True:
        print("Finished in {} seconds".format(time.time() - start))
        return iplot(fig)

    else:
        print("Finished in {} seconds".format(time.time() - start))
        return plot(fig, filename=filename)



from typing import Optional, Union, Any
import numpy as np
import random
from collections import Counter
import itertools

import navis
import neuprint as neu
import navis.interfaces.neuprint as nvneu
import pymaid

import scipy.sparse as sparse
from scipy.sparse.linalg import spsolve
from sklearn.cluster import DBSCAN
import phate
import scprep
import scipy.spatial.distance as ssd

from neuroboom import utils as nbu


def prepare_neuron(x,
change_units = True,
factor = 1e3):

    if isinstance(x, (navis.TreeNeuron, navis.NeuronList)):

        if isinstance(x, navis.NeuronList):

            if len(x) > 1:

                raise ValueError('Need to pass a SINGLE neuron')

            else:

                x = x[0]

        node_sort = dict([(i, k) for i, k in zip(range(len(x.nodes)), navis.graph_utils.node_label_sorting(x))])
        node_sort_rev = {i : j for j, i in node_sort.items()}
        navis.downsample_neuron(x, downsampling_factor = float('inf'), inplace = True)
        x.nodes['node_rank'] = x.nodes.node_id.map(node_sort_rev).tolist()
        x.nodes.sort_values(by = ['node_rank'], ascending = True, inplace = True)
        x.nodes.reset_index(drop = True, inplace = True)

        #x = navis_calc_cable(x, return_skdata=True)
        x = nbu.calc_cable(x, return_skdata=True)

        if not change_units:

            return(x)

        else:

            x.nodes['x'] = x.nodes['x'] / factor
            x.nodes['y'] = x.nodes['y'] / factor
            x.nodes['z'] = x.nodes['z'] / factor
            x.nodes['radius'] = x.nodes['radius'] / factor
            x.nodes['parent_dist'] = x.nodes['parent_dist'] / factor

            return(x)


    elif isinstance(x, (pymaid.CatmaidNeuron, pymaid.CatmaidNeuronList)):

        if isinstance(x, pymaid.CatmaidNeuronList):

            if len(x) > 1:

                raise ValueError('Need to pass a SINGE neurno')

            else:

                x = x[0]

        node_sort = dict([(i, k) for i, k in zip(range(len(x.nodes)), pymaid.node_label_sorting(x))])
        node_sort_rev = {i : j for j, i in node_sort.items()}
        x = pymaid.downsample_neuron(x, resampling_factor = float('inf'))
        x = pymaid.guess_radius(x)

        x.nodes['node_rank'] = x.nodes.treenode_id.map(node_sort_rev).tolist()
        x.nodes.sort_values(by = ['node_rank'], ascending = True, inplace = True)
        x.nodes.reset_index(drop = True, inplace = True)

        x = pymaid.calc_cable(x, return_skdata=True)

        if not change_units:

            return(x)

        else:

            x.nodes['x'] = x.nodes['x'] / factor
            x.nodes['y'] = x.nodes['y'] / factor
            x.nodes['z'] = x.nodes['z'] / factor
            x.nodes['radius'] = x.nodes['radius'] / factor
            x.nodes['parent_dist'] = x.nodes['parent_dist'] / factor

            return(x)

    else:

        raise ValueError('Need to pass either a Navis or a Catmaid neuron type!')


def calculate_M_mat(x, Rm, Ra, Cm, solve = False):

    if isinstance(x, (navis.TreeNeuron, navis.NeuronList)):

        if isinstance(x, navis.NeuronList):

            if len(x) > 1:

                raise ValueError('Need to pass a SINGLE neuron')

            else:

                x = x[0]

        tn_to_index = dict(zip(x.nodes.node_id, x.nodes.index))
        nofcomps = x.nodes.shape[0]
        M = np.zeros((nofcomps, nofcomps))

        compdiam = x.nodes.radius * 2
        complength = np.zeros(nofcomps)

        # skip root node

        for i in range(1, nofcomps):

            aind = int(x.nodes.node_id[i])
            bind = int(x.nodes.parent_id[i])

            axyz = x.nodes[x.nodes.node_id == aind][['x','y','z']].values
            bxyz = x.nodes[x.nodes.node_id == bind][['x','y','z']].values

            complength[i] = np.sqrt(np.sum((axyz - bxyz) ** 2))

            meandiam = (x.nodes[x.nodes.node_id == aind].radius.values + x.nodes[x.nodes.node_id == bind].radius.values) * .5

            area = (meandiam ** 2) / (4.0 * np.pi)

            M[ tn_to_index[bind], tn_to_index[aind]] = -area / complength[i] / Ra * 10**(-4)
            M[ tn_to_index[aind], tn_to_index[bind]] = M[ tn_to_index[bind], tn_to_index[aind]]

        complength[0] = complength[1]

        gleak = (compdiam * np.pi * complength) / (Rm * 10 ** 8)
        memcap = (compdiam * np.pi * complength) * Cm * (10 ** -6) / (10 ** 8)

        for i in range(nofcomps):
            M[i, i] = gleak[i] - np.sum(M[i])

        M = sparse.csr_matrix(M)

        if solve:

            M_solved = np.linalg.inv(sparse.csr_matrix.todense(M))

            return(M_solved, memcap)

        return(M, memcap)

    elif isinstance(x, (pymaid.CatmaidNeuron, pymaid.CatmaidNeuronList)):

        if isinstance(x, pymaid.CatmaidNeuronList):

            if len(x) > 1:

                raise ValueError('Need to pass a SINGLE neuron')

            else:

                x = x[0]

        tn_to_index = dict(zip(x.nodes.treenode_id, x.nodes.index))
        nofcomps = x.nodes.shape[0]
        M = np.zeros((nofcomps, nofcomps))

        compdiam = x.nodes.radius * 2
        complength = np.zeros(nofcomps)

        # skip root node

        for i in range(1, nofcomps):

            aind = int(x.nodes.treenode_id[i])
            bind = int(x.nodes.parent_id[i])

            axyz = x.nodes[x.nodes.treenode_id == aind][['x','y','z']].values
            bxyz = x.nodes[x.nodes.treenode_id == bind][['x','y','z']].values

            complength[i] = np.sqrt(np.sum((axyz - bxyz) ** 2))

            meandiam = (x.nodes[x.nodes.treenode_id == aind].radius.values + x.nodes[x.nodes.treenode_id == bind].radius.values) * .5

            area = (meandiam ** 2) / (4.0 * np.pi)

            M[ tn_to_index[bind], tn_to_index[aind]] = -area / complength[i] / Ra*10**(-4)
            M[ tn_to_index[aind], tn_to_index[bind]] = M[ tn_to_index[bind], tn_to_index[aind]]

        complength[0] = complength[1]

        gleak = (compdiam * np.pi * complength) / (Rm * 10 ** 8)
        memcap = (compdiam * np.pi * complength) * Cm*(10 ** -6) / (10 ** 8)

        for i in range(nofcomps):
            M[i, i] = gleak[i] - np.sum(M[i])

        M = sparse.csr_matrix(M)

        if solve:

            M_solved = np.linalg.inv(sparse.csr_matrix.todense(M))
            return(M_solved, memcap)

        return(M, memcap)

    else:

        raise ValueError('Need to pass either a Navis or a Catmaid neuron type!')


def current_injection(conducM, curramp):

    nofcomps = conducM.shape[0]
    Vm_mat = np.zeros((nofcomps, nofcomps))

    #currinj = np.zeros(nofcomps)
    #currinj[injection_comp] = curramp

    #Vm = spsolve(conducM, curr_inj)
    #Vm = 1e3 * Vm #in mV

    for i in range(0, nofcomps):

        currinj = np.zeros(nofcomps)
        currinj[i] = curramp

        Vm = spsolve(conducM, currinj)
        #Vm = 1e3 * Vm # in mV

        Vm_mat[i,:] = Vm

    return(Vm_mat)



def dbs_func(x, eps: float = 1.0, min_samples: int = 10):

    """
    Parameters
    --------
    x:              A matrix to perform DBSCAN on

    eps:            float
                    The radius of the search circle

    min_samples:    int
                    The minimum number of samples to include in a cluster

    Returns
    --------
    labels:         cluster label of each datapoint
    n_clusters_:    number of unique clusters
    n_noise_:       number of noise points

    Examples
    --------

    """

    db = DBSCAN(eps=eps, min_samples=min_samples).fit(x)
    labels = db.labels_
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    n_noise_ = list(labels).count(-1)

    return (labels, n_clusters_, n_noise_)

def find_clusters(x, phate_operator, eps: float = 1.0, min_samples: int = 10):

    """
    Parameters
    --------
    x:              A matrix to perform DBSCAN on

    phate_operator: phate.operator object
                    The operator used to perform the dimensionality reduction
                    The number of dimensions to reduce to
                    is specified in this operator object

    eps:            float
                    Radius of the search circle

    min_samples:    int
                    Minimum number of samples to include in a cluster

    Returns
    --------
    mat_red:        Matrix wih dimensions reduced
    labels:         Cluster label of each datapoint
    n_clusters_:    Number of unique clusters
    n_noise_:       Number of noise points

    Examples
    --------

    """

    # performing PHATE dimensionality reduction
    mat_red = phate_operator.fit_transform(x)

    # performing DBSCAN
    labels, n_clusters_, n_noise_ = dbs_func(mat_red, eps, min_samples)

    return (mat_red, labels, n_clusters_, n_noise_)


def match_connectors_to_nodes(synapse_connections, neuron, synapse_type = 'post'):

    if synapse_type == 'post':

        u = synapse_connections[['x_post','y_post','z_post']].values
        v = neuron.postsynapses[['x','y','z']].values

        data = ssd.cdist(u, v, metric = 'euclidean')

        dist_mat = pd.DataFrame(index = [i for i in range(0, synapse_connections.shape[0])],
                                columns = [i for i in range(0, neuron.postsynapses.shape[0])],
                                data = data)



        ind = [np.argmin(dist_mat.iloc[i, :]) for i in range(0, synapse_connections.shape[0])]

        syn_con = synapse_connections.copy()
        syn_con['connector'] = [neuron.postsynapses.iloc[i, :].connector_id for i in ind]

        # creating a connector to node dictionary

        c2n = dict(zip(neuron.postsynapses.connector_id.tolist(), neuron.postsynapses.node_id.tolist()))
        syn_con['node'] = syn_con.connector.map(c2n)

    elif synapse_type == 'pre':

        u = synapse_connections[['x_pre','y_pre','z_pre']].values
        v = neuron.presynapses[['x','y','z']].values

        data = ssd.cdist(u, v, metric = 'euclidean')

        dist_mat = pd.DataFrame(index = [i for i in range(0, synapse_connections.shape[0])],
                                columns = [i for i in range(0, neuron.presynapses.shape[0])],
                                data = data)



        ind = [np.argmin(dist_mat.iloc[i, :]) for i in range(0, synapse_connections.shape[0])]

        syn_con = synapse_connections.copy()
        syn_con['connector'] = [neuron.presynapses.iloc[i, :].connector_id for i in ind]

        # creating connector to node dictionary

        c2n = dict(zip(neuron.presynapses.connector_id.tolist(), neuron.presynapses.node_id.tolist()))
        syn_con['node'] = syn_con.connector.map(c2n)


    return(syn_con)

def permute_start_end(test, test_in_roi, cluster):

    start_end = [i for i in itertools.permutations(test.nodes[test_in_roi][test.nodes[test_in_roi].node_cluster == cluster].node_id.tolist(), 2)]

    return(start_end)

def cluster_to_all_nodes(neuron, start_end_node_pairs):

    g = nbd.create_graph_structure(neuron, returned_object = 'graph')
    g_rev = g.reverse()

    nodes_of_cluster = []

    for i, j in start_end_node_pairs:

        paths = nx.all_simple_paths(g_rev, source = i, target = j)

        for path in paths:

            nodes_of_cluster.append(path)

    nodes_of_cluster = list(np.unique(list(chain.from_iterable(nodes_of_cluster))))
    return(nodes_of_cluster)

def quick_optimisation(x, min_samples_start):

    """
    x : a neuronlist
    min_samples_start: the minimum number of samples to start off with

    """
    min_samples_optim_values = []
    err = []

    for i in x:

        print(i.id, i.name)

        n = i.copy()

        try:

            n_prep = prepare_neuron(n, change_units = True, factor = 1e3)

        except IndexError:

            err.append(i.id)

            continue

        n_m, n_memcap = calculate_M_mat(n_prep, solve = False)
        n_m_solved = np.linalg.inv(sparse.csr_matrix.todense(n_m))

        phate_operator = phate.PHATE(n_components = 3, n_jobs = -2, verbose = False)

        min_sample_range = [i for i in range(0, min_samples_start + 1)][::-1]

        for j in min_sample_range:

            mat_red, labels, n_clust, n_noise = nbm.find_clusters(n_m_solved, phate_operator, eps=1e-02, min_samples = j)

            if n_noise == 0:

                min_samples_optim_values.append((i.id, j))

                break

            else:

                continue

    return(min_samples_optim_values, err)

def find_compartments_in_roi(neuron, roi, min_samples):

    n = neuron.copy()
    n_prep = prepare_neuron(n, change_units=True, factor = 1e3)
    n_m, n_memcap = calculate_M_mat(n_prep, solve = False)
    n_m_solved = np.linalg.inv(sparse.csr_matrix.todense(n_m))

    phate_operator = phate.PHATE(n_components=3, n_jobs=-2, verbose = False)

    mat_red, labels, n_clusters, n_noise = nbm.find_clusters(n_m_solved, phate_operator, eps = 1e-02, min_samples = min_samples)

    node_to_label = dict(zip(range(0, len(labels)), labels))

    n_prep.nodes['node_cluster'] = n_prep.nodes.index.map(node_to_label)

    n_in_roi = navis.in_volume(n_prep.nodes[['x','y','z']] * 1e3, roi, mode = 'IN', inplace = False)

    return(n_prep, n_in_roi)

def matching_inputs_to_compartments(neuron_id, roi):

    # Fetch the healed skeleton
    full_skeleton = nvneu.fetch_skeletons(neuron_id, heal = True)[0]
    # which of the whole neuron's nodes are in the roi
    skeleton_in_roi = navis.in_volume(full_skeleton.nodes[['x','y','z']], roi, mode='IN')

    # Fetch the neurons synapses
    postsyn = nvneu.fetch_synapse_connections(target_criteria = neuron_id)
    # match the connectors to nodes
    syn_to_node = match_connectors_to_nodes(postsyn, full_skeleton, synapse_type = 'post')
    # which of these are in the roi
    roi_syn_con = syn_to_node[syn_to_node.node.isin(full_skeleton.nodes[skeleton_in_roi].node_id.tolist())].copy()

    # Count how many synapses the upstream neurons have on your neuron of interest
    n_syn = dict(Counter(roi_syn_con.bodyId_pre.tolist()).most_common())

    # fetch these neurons
    a, b = nvneu.fetch_neurons(roi_syn_con.bodyId_pre.unique())
    a['n_syn'] = [n_syn[i] for i in a.bodyId.tolist()]

    # adding the instance names to the synapses in the roi
    bid_to_instance = dict(zip(a.bodyId.tolist(), a.instance.tolist()))
    roi_syn_con['instance'] = [bid_to_instance[i] for i in roi_syn_con.bodyId_pre]

    # compartmentalise the neuron and find the nodes in each compartment for the prepared neuron
    compartments_in_roi, nodes_in_roi = find_compartments_in_roi(full_skeleton, ca, 6)
    clusters = compartments_in_roi.nodes.node_cluster.unique()

    cluster_dict = {}

    # find the nodes that make up each compartment in the full neuron
    for i in clusters:

        clust_nodes = cluster_to_all_nodes(full_skeleton, start_end_node_pairs=permute_start_end(compartments_in_roi, nodes_in_roi, cluster = i))
        cluster_dict[i] = clust_nodes

    #cluster_dict = {k : s for s, k in cluster_dict.items()}
    #roi_syn_con['compartment'] = [cluster_dict[i] for i in roi_syn_con.node.tolist()]

    return(cluster_dict)
    #return(roi_syn_con)

    #return(full_skeleton, skeleton_in_roi, skeleton_prep, roi_syn_con, a)


def find_compartments_of_missing_nodes(roi_syn_con, nodes_with_cluster, full_neuron, ds_neuron):

    """
    DS neuron -- downsampled_neuron
    """

    # find the connections associated to nodes that do not have a cluster
    nodes_to_query = roi_syn_con[~roi_syn_con.node.isin(nodes_with_cluster)].node.tolist()

    # create a geodesic matrix between the nodes that do not have a compartment and the original, non-downsampled neuron
    gmat = navis.geodesic_matrix(full_neuron, nodes_to_query)

    # subset the above geodesic matrix columns to only contain the nodes in the downsampled neuron
    reduced_gmat = gmat.T[np.isin(gmat.T.index, ds_neuron.nodes.node_id.tolist())].T.copy()

    # find the node which is the smallest distance away from the unclassified node
    closest_classified_node = list(reduced_gmat.columns[np.argmin(reduced_gmat.values, axis = 1)])

    closest_classified_node_to_cluster = ds_neuron.nodes[ds_neuron.nodes.node_id.isin(np.unique(closest_classified_node))]
    closest_classified_node_to_cluster = dict(zip(closest_classified_node_to_cluster.node_id, closest_classified_node_to_cluster.node_cluster))

    query_nodes_to_closest_node = dict(zip(np.unique(nodes_to_query), np.array(reduced_gmat.columns[np.argmin(reduced_gmat.values, axis = 1)])))

    comps = [closest_classified_node_to_cluster[query_nodes_to_closest_node[i]] for i in query_nodes_to_closest_node.keys()]

    query_nodes_to_compartment = dict(zip(query_nodes_to_closest_node.keys(), comps))

    return(query_nodes_to_compartment)

def compartmentalise_neuron(neuron_id, roi):

    # Fetching the neuron
    ds_neuron = nvneu.fetch_skeletons(neuron_id, heal = True)[0]
    original_neuron = nvneu.fetch_skeletons(neuron_id, heal = True)[0]

    # Electrotonic model
    DS_NEURON = prepare_neuron(ds_neuron, change_units = True, factor = 1e3)
    test_m, test_memcap = calculate_M_mat(DS_NEURON, solve = False)
    test_m_solved = np.linalg.inv(sparse.csr_matrix.todense(test_m))

    # running PHATE
    phate_operator = phate.PHATE(n_components = 3, n_jobs = -2, verbose = False)

    mat_red, labels, n_clusters, n_noise = nbm.find_clusters(test_m_solved,
                                                              phate_operator,
                                                              eps = 1e-02,
                                                              min_samples = 6)

    # index and labels
    index_to_label = dict(zip(range(0, len(labels)), labels))
    ds_neuron.nodes['node_cluster'] = ds_neuron.nodes.index.map(index_to_label)

    node_to_compartment = dict(zip(ds_neuron.nodes.node_id.tolist(),
                                   ds_neuron.nodes.node_cluster.tolist()))

    unique_compartments = ds_neuron.nodes.node_cluster.unique()

    # Finding the cluster of the nodes that were removed when downsampling the neuron
    whole_neuron_node_to_cluster = []

    for i in unique_compartments:

        start_end = [i for i in itertools.permutations(ds_neuron.nodes[ds_neuron.nodes.node_cluster == i].node_id.tolist(), 2)]

        nodes_of_cluster = cluster_to_all_nodes(original_neuron, start_end)

        node_to_cluster_dictionary = dict(zip(nodes_of_cluster, [i] * len(nodes_of_cluster)))

        whole_neuron_node_to_cluster.append(node_to_cluster_dictionary)

    whole_neuron_node_to_cluster_dict = {k : v for d in whole_neuron_node_to_cluster for k, v in d.items()}

    #Fetching postsynapses
    ds_neuron_postsynapses = nvneu.fetch_synapse_connections(target_criteria=neuron_id)

    ds_neuron_synapse_to_node = match_connectors_to_nodes(ds_neuron_postsynapses,
                                                          original_neuron,
                                                          synapse_type = 'post')

    #Which nodes are in the CA?
    skeleton_in_roi = navis.in_volume(original_neuron.nodes[['x','y','z']].values, roi, inplace = False)
    roi_syn_con = ds_neuron_synapse_to_node[ds_neuron_synapse_to_node.node.isin(original_neuron.nodes[skeleton_in_roi].node_id.tolist())].copy()

    a, b = nvneu.fetch_neurons(roi_syn_con.bodyId_pre.unique())
    bid_to_instance = dict(zip(a.bodyId.tolist(), a.instance.tolist()))
    roi_syn_con['instance'] = [bid_to_instance[i] for i in roi_syn_con.bodyId_pre]

    nodes_to_query = roi_syn_con[~roi_syn_con.node.isin(whole_neuron_node_to_cluster_dict.keys())].node.tolist()
    comp_of_missing_nodes = find_compartments_of_missing_nodes(roi_syn_con,
                                                               list(whole_neuron_node_to_cluster_dict.keys()),
                                                               original_neuron,
                                                               ds_neuron)

    whole_neuron_node_to_cluster_dict = {**whole_neuron_node_to_cluster_dict, **comp_of_missing_nodes}

    roi_syn_con['compartment'] = [whole_neuron_node_to_cluster_dict[i] for i in roi_syn_con.node.tolist()]

    original_neuron = node_to_compartment_full_neuron(original_neuron, ds_neuron)

    return(original_neuron, ds_neuron, roi_syn_con)



# INSERT LICENSE

# This script contains statistical functions
import copy
import random
from collections import Counter
from typing import Optional, Union

import navis
import navis.interfaces.neuprint as nvneu
import numpy as np
import pandas as pd
from scipy import stats
from tqdm import tqdm

from neuroboom.utils import check_valid_neuron_input


def presynapse_focality(
    x: Union[navis.TreeNeuron, navis.NeuronList],
    heal_fragmented_neuron: bool = False,
    confidence_threshold: tuple((float, float)) = (0.9, 0.9),
    num_synapses_threshold: int = 1,
):

    """
    Finds the connections that are downstream of 'x', where the presynpases of 'x' are focalised

    Parameters
    --------
    x:                       A matrix to perform DBSCAN on

    heal_fragmented_neuron:  bool
                             Whether to heal the neuron or not.
                             N.B. Its better to heal neurons during
                             import to save time in this function.

    connector_confidence:    tuple of floats
                             The confidence value used to threshold the synapses.
                             The first value (connector_confidence[0]) will be used to threshold presynapses
                             The second value (connector_confidence[1]) will be used to threshold postsynapses

    num_samples_threshold:   int
                             The minimum number of synapses a partner must have
                             to be included in the permutation test

    Returns
    --------
    synapse_connections:     A dataframe detailing the presynaptic connections
    df:                      A dataframe to be populated by the permutation test function

    Examples
    --------

    """

    x = check_valid_neuron_input(x)

    if heal_fragmented_neuron is True:

        x = navis.heal_fragmented_neuron(x)

    # Getting the connector table of synapses where x.id is the source
    synapse_connections = navis.interfaces.neuprint.fetch_synapse_connections(
        source_criteria=x.id
    )
    synapse_connections.astype(object)
    synapse_connections["node_id"] = object

    for i, j in tqdm(
        enumerate(synapse_connections[["x_pre", "y_pre", "z_pre"]].values)
    ):

        equals = np.equal(j, x.presynapses[["x", "y", "z"]].values)
        all_equals = [np.all(k) for k in equals]
        node_id = x.presynapses[all_equals].node_id.tolist()
        synapse_connections.at[i, "node_id"] = node_id

    truth_list = [
        True if len(np.unique(i)) > 1 else False
        for i in synapse_connections.node_id.values
    ]
    if synapse_connections[truth_list].shape[0] == 0:

        synapse_connections.node_id = [
            np.unique(k)[0] for k in synapse_connections.node_id.tolist()
        ]

    else:

        return "There are synapses associated with multiple nodes!!!!"

    synapse_connections = synapse_connections[
        synapse_connections.confidence_pre > confidence_threshold[0]
    ][synapse_connections.confidence_post > confidence_threshold[1]].copy()
    count = Counter(synapse_connections.bodyId_post.tolist())
    count = {
        k: v for k, v in sorted(count.items(), key=lambda item: item[1], reverse=True)
    }
    truth_list = [
        True if count[i] > num_synapses_threshold else False
        for i in synapse_connections.bodyId_post
    ]
    synapse_connections = synapse_connections[truth_list].copy()

    df = pd.DataFrame()
    df["partner_neuron"] = synapse_connections.bodyId_post.unique().tolist()
    df["gT"] = ""
    df["significance_val"] = ""
    df["p_val"] = ""
    df["num_syn"] = [count[i] for i in df.partner_neuron]

    return (synapse_connections, df)


def postsynapse_focality(
    x: Union[navis.TreeNeuron, navis.NeuronList],
    heal_fragmented_neuron: bool = False,
    split_neuron: bool = False,
    confidence_threshold: tuple((float, float)) = (0.9, 0.9),
    num_synapses_threshold: int = 1,
):

    """
    Finds the connections that are downstream of 'x', where the presynpases of 'x' are focalised

    Parameters
    --------
    x:                       A matrix to perform DBSCAN on

    heal_fragmented_neuron:  bool
                             Whether to heal the neuron or not.
                             N.B. Its better to heal neurons during
                             import to save time in this function.

    connector_confidence:    tuple of floats
                             The confidence value used to threshold the synapses.
                             The first value (connector_confidence[0]) will be used to threshold presynapses
                             The second value (connector_confidence[1]) will be used to threshold postsynapses

    num_samples_threshold:   int
                             The minimum number of synapses a partner must have
                             to be included in the permutation test

    Returns
    --------
    synapse_connections:     A dataframe detailing the presynaptic connections
    df:                      A dataframe to be populated by the permutation test function

    Examples
    --------

    """

    x = check_valid_neuron_input(x)

    if heal_fragmented_neuron is True:

        x = navis.heal_fragmented_neuron(x)

    # Getting the connector table of synapses where x.id is the source
    synapse_connections = navis.interfaces.neuprint.fetch_synapse_connections(
        target_criteria=x.id
    )
    synapse_connections.astype(object)
    synapse_connections["node_id"] = object

    for i, j in tqdm(
        enumerate(synapse_connections[["x_post", "y_post", "z_post"]].values)
    ):

        node_id = x.postsynapses[
            [np.all(k) for k in np.equal(j, x.postsynapses[["x", "y", "z"]].values)]
        ].node_id.tolist()
        synapse_connections.at[i, "node_id"] = node_id

    the_truth = [
        True if len(np.unique(i)) > 1 else False
        for i in synapse_connections.node_id.values
    ]

    if synapse_connections[the_truth].shape[0] == 0:

        synapse_connections.node_id = [
            np.unique(k)[0] for k in synapse_connections.node_id.tolist()
        ]

    else:
        return "There are synapses associated with multiple nodes!!!!"

    synapse_connections = synapse_connections[
        synapse_connections.confidence_pre > confidence_threshold[0]
    ][synapse_connections.confidence_post > confidence_threshold[1]].copy()
    count = Counter(synapse_connections.bodyId_pre.tolist())
    count = {
        k: v for k, v in sorted(count.items(), key=lambda item: item[1], reverse=True)
    }
    truth_list = [
        True if count[i] > num_synapses_threshold else False
        for i in synapse_connections.bodyId_pre
    ]
    synapse_connections = synapse_connections[truth_list].copy()

    df = pd.DataFrame()
    df["partner_neuron"] = synapse_connections.bodyId_pre.unique().tolist()
    df["gT"] = ""
    df["significance_val"] = ""
    df["p_val"] = ""
    df["num_syn"] = [count[i] for i in df.partner_neuron]

    return (synapse_connections, df)


def permut_test(
    x: Union[navis.TreeNeuron, navis.NeuronList],
    measuring_node: int,
    synapse_connections: pd.DataFrame,
    relation: str = "presyn",
    num_iter: int = 10,
    df: Optional = None,
    count: Optional = None,
):

    """
    Runs a permutation test on the geodesic distances for connections

    Parameters
    --------
    x:                       navis.TreeNeuron

    measuring_node:          int
                             Node ID for which to measure the geodesic distance of synapses

    synapse_connections:     pandas.DataFrame
                             A DataFrame containin the synaptic connections
                             upon which the permutation test will be executed.

    relation:                str
                             Whether the synaptic connections included in
                             synapse_connections are presynapses or postsynapses

    num_iter:                int
                             Number of iterations to run the permutation test for

    df:                      pandas.DataFrame
                             A DataFrame to record the p_values of the permutation test

    count:                   dict
                             The total number of synaptic connections the
                             partner neuron has onto x


    Returns
    --------
    df:                      A dataframe populated by the permutation test function
                             containing the p_values of the partners, determined by
                             the locations of their connections.


    Examples
    --------

    """

    x = check_valid_neuron_input(x)

    geo_mat = navis.geodesic_matrix(x, node_ids=measuring_node)
    geo_mat = geo_mat.T
    geo_mat.sort_values(by=[measuring_node], ascending=True, inplace=True)

    for k, j in tqdm(enumerate(df.partner_neuron)):

        if relation == "presyn":

            total_distribution = geo_mat[
                ~geo_mat.index.isin(
                    synapse_connections[
                        synapse_connections.bodyId_post == j
                    ].node_id.tolist()
                )
            ]
            specific_distribution = geo_mat[
                geo_mat.index.isin(
                    synapse_connections[
                        synapse_connections.bodyId_post == j
                    ].node_id.tolist()
                )
            ]

        elif relation == "postsyn":

            total_distribution = geo_mat[
                ~geo_mat.index.isin(
                    synapse_connections[
                        synapse_connections.bodyId_pre == j
                    ].node_id.tolist()
                )
            ]
            specific_distribution = geo_mat[
                geo_mat.index.isin(
                    synapse_connections[
                        synapse_connections.bodyId_pre == j
                    ].node_id.tolist()
                )
            ]

        total_mean = np.average(total_distribution.values)
        specific_mean = np.average(specific_distribution.values)

        gT = np.abs(total_mean - specific_mean)
        df.iloc[k, 1] = gT

        pV = list(total_distribution.values) + list(specific_distribution.values)
        pV = [i[0] for i in pV]

        pS = copy.copy(pV)
        pD = []

        for i in range(0, num_iter):

            random.shuffle(pS)
            pD.append(
                np.abs(
                    np.average(pS[0: int(len(pS) / 2)])
                    - np.average(pS[int(len(pS) / 2):])
                )
            )

        p_val = len(np.where(pD >= gT)[0]) / num_iter
        df.iloc[k, 2] = p_val

        ttest = stats.ttest_ind(total_distribution.values, specific_distribution.values)
        df.iloc[k, 3] = ttest[1][0]

    df.sort_values(by=["p_val"], ascending=True, inplace=True)
    df.reset_index(inplace=True)
    return df


def permutation_test_complete(
    x: Union[navis.TreeNeuron, navis.NeuronList],
    n_iter: int = 10,
    remove_fragments: bool = True,
    confidence_threshold: tuple = (0.9, 0.9)
):

    """
    A wrapper function for the presynaptic and postsynaptic permutation test
    functions, so that both can be performed with minimal code writing
    by the user

    Parameters
    --------
    x:                       navis.TreeNeuron

    n_iter:                  int
                             Number of iterations to run the permutation test for

    remove_fragments:        bool
                             Whether to remove partners that are fragments/
                             have not been traced to completion/
                             do not have a soma.


    Returns
    --------
    presyn_pt:               A dataframe populated by the permutation test function
                             containing the p_values of the presynaptic partners,
                             determined by the locations of their connections.

    postsyn_pt:              A dataframe populated by the permutation test function
                             containing the p_values of the postsynaptic partners,
                             determined by the locations of their connections.


    Examples
    --------

    """

    a_pre, a_df = presynapse_focality(
        x, heal_fragmented_neuron=False, confidence_threshold=confidence_threshold
    )
    b_post, b_df = postsynapse_focality(
        x, heal_fragmented_neuron=False, confidence_threshold=confidence_threshold
    )

    presyn_pt = permut_test(
        x,
        x.root[0],
        synapse_connections=a_pre,
        relation="presyn",
        num_iter=n_iter,
        df=a_df,
    )
    postsyn_pt = permut_test(
        x,
        x.root[0],
        synapse_connections=b_post,
        relation="postsyn",
        num_iter=n_iter,
        df=b_df,
    )

    a_partner_neurons, a_roi = nvneu.fetch_neurons(presyn_pt.partner_neuron.tolist())
    b_partner_neurons, b_roi = nvneu.fetch_neurons(postsyn_pt.partner_neuron.tolist())

    a_type_dict = dict(
        zip(a_partner_neurons.bodyId.tolist(), a_partner_neurons.type.tolist())
    )
    b_type_dict = dict(
        zip(b_partner_neurons.bodyId.tolist(), b_partner_neurons.type.tolist())
    )

    partner_dict = {**a_type_dict, **b_type_dict}

    presyn_pt["partner_type"] = presyn_pt.partner_neuron.map(
        partner_dict, na_action="ignore"
    )
    postsyn_pt["partner_type"] = postsyn_pt.partner_neuron.map(
        partner_dict, na_action="ignore"
    )

    if remove_fragments:

        presyn_pt = presyn_pt[~presyn_pt.partner_type.isnull()].copy()
        postsyn_pt = postsyn_pt[~postsyn_pt.partner_type.isnull()].copy()

    return (presyn_pt, postsyn_pt)



# INSERT LICENSE
import random
from typing import Optional, Union, Any

import numpy as np
import scipy.sparse as sparse
from scipy.sparse.linalg import spsolve
import scipy.spatial.distance as ssd
from sklearn.cluster import DBSCAN

import navis
import pymaid
import seaborn as sns

import itertools
from collections import Counter


from neuroboom import dendrogram as nbd
from neuroboom.utils import calc_cable, check_valid_neuron_input, check_valid_pymaid_input

# Contains functions to model steady state passive electrotonic properties of neurons
def prepare_neuron(
    x: Union[
        navis.TreeNeuron,
        navis.NeuronList,
        pymaid.core.CatmaidNeuron,
        pymaid.core.CatmaidNeuronList,
    ],
    change_units: bool = True,
    factor: int = 1e3,
):

    """
    Takes a navis or pymaid neuron and prepares it for electrotonic modelling

    Paramters
    ---------
    x:                a pymaid/Catmaid neuron object


    return_skdata:    bool
                    whether to return a list of node_ids topologically sorted
                    or to return a dict where the keys are treenodes and the values
                    are ranking in the topological sort

    factor:           the factor to scale the nodes from um to nm

    Returns
    --------

    x:                the neuron rescaled and ordered for electro-modelling

    Examples
    --------



    """

    if check_valid_neuron_input(x):

        node_sort = dict(
            [(i, k) for i, k in zip(range(len(x.nodes)), navis.graph_utils.node_label_sorting(x))]
        )
        node_sort_rev = {i: j for j, i in node_sort.items()}
        navis.downsample_neuron(x, downsampling_factor=float("inf"), inplace=True)
        x.nodes["node_rank"] = x.nodes.node_id.map(node_sort_rev).tolist()
        x.nodes.sort_values(by=["node_rank"], ascending=True, inplace=True)
        x.nodes.reset_index(drop=True, inplace=True)

        x = calc_cable(x, return_skdata=True)

        if not change_units:

            return x

        else:

            x.nodes["x"] = x.nodes["x"] / factor
            x.nodes["y"] = x.nodes["y"] / factor
            x.nodes["z"] = x.nodes["z"] / factor
            x.nodes["radius"] = x.nodes["radius"] / factor
            x.nodes["parent_dist"] = x.nodes["parent_dist"] / factor

            return x

    elif check_valid_pymaid_input(x):

        node_sort = dict(
            [(i, k) for i, k in zip(range(len(x.nodes)), pymaid.node_label_sorting(x))]
        )
        node_sort_rev = {i: j for j, i in node_sort.items()}
        x = pymaid.downsample_neuron(x, resampling_factor=float("inf"))
        x = pymaid.guess_radius(x)

        x.nodes["node_rank"] = x.nodes.treenode_id.map(node_sort_rev).tolist()
        x.nodes.sort_values(by=["node_rank"], ascending=True, inplace=True)
        x.nodes.reset_index(drop=True, inplace=True)

        x = pymaid.calc_cable(x, return_skdata=True)

        if not change_units:

            return x

        else:

            x.nodes["x"] = x.nodes["x"] / factor
            x.nodes["y"] = x.nodes["y"] / factor
            x.nodes["z"] = x.nodes["z"] / factor
            x.nodes["radius"] = x.nodes["radius"] / factor
            x.nodes["parent_dist"] = x.nodes["parent_dist"] / factor

            return x

    else:

        raise ValueError("Need to pass either a Navis or a Catmaid neuron type!")


def calculate_M_mat(
    x: Union[
        navis.TreeNeuron,
        navis.NeuronList,
        pymaid.core.CatmaidNeuron,
        pymaid.core.CatmaidNeuronList,
    ],
    Ra: float = 266.1,
    Rm: float = 20.8,
    Cm: float = 0.8,
    solve: bool = False,
):

    """
    Calculates the conductance matrix for a given neuron

    Paramters
    ---------
    x:                  a pymaid/Catmaid neuron object
                        This object needs to be passed through prepare_neuron()
                        before running this function


    Ra:                 float
                        Axial resistance with units in ohms / cm

    Rm:                 float
                        Membrane resistance with units in kilo-ohms / cm^2

    Cm:                 float
                        Membrane capacitance with units in uF / cm^2
                        uF = microFarads

    solve:              bool
                        Whether to solve the matrix (calculate its inverse)

    Returns
    --------

    M:
    M_solved:
    memcap:

    Examples
    --------



    """

    if check_valid_neuron_input(x):

        tn_to_index = dict(zip(x.nodes.node_id, x.nodes.index))
        nofcomps = x.nodes.shape[0]
        M = np.zeros((nofcomps, nofcomps))

        compdiam = x.nodes.radius * 2
        complength = np.zeros(nofcomps)

        # skip root node

        for i in range(1, nofcomps):

            aind = int(x.nodes.node_id[i])
            bind = int(x.nodes.parent_id[i])

            axyz = x.nodes[x.nodes.node_id == aind][["x", "y", "z"]].values
            bxyz = x.nodes[x.nodes.node_id == bind][["x", "y", "z"]].values

            complength[i] = np.sqrt(np.sum((axyz - bxyz) ** 2))

            meandiam = (
                x.nodes[x.nodes.node_id == aind].radius.values
                + x.nodes[x.nodes.node_id == bind].radius.values
            ) * 0.5

            area = (meandiam ** 2) / (4.0 * np.pi)

            M[tn_to_index[bind], tn_to_index[aind]] = (
                -area / complength[i] / Ra * 10 ** (-4)
            )
            M[tn_to_index[aind], tn_to_index[bind]] = M[
                tn_to_index[bind], tn_to_index[aind]
            ]

        complength[0] = complength[1]

        gleak = (compdiam * np.pi * complength) / (Rm * 10 ** 8)
        memcap = (compdiam * np.pi * complength) * Cm * (10 ** -6) / (10 ** 8)

        for i in range(nofcomps):
            M[i, i] = gleak[i] - np.sum(M[i])

        M = sparse.csr_matrix(M)

        if solve:

            M_solved = np.linalg.inv(sparse.csr_matrix.todense(M))

            return (M_solved, memcap)

        return (M, memcap)

    elif check_valid_pymaid_input(x):

        tn_to_index = dict(zip(x.nodes.treenode_id, x.nodes.index))
        nofcomps = x.nodes.shape[0]
        M = np.zeros((nofcomps, nofcomps))

        compdiam = x.nodes.radius * 2
        complength = np.zeros(nofcomps)

        # skip root node

        for i in range(1, nofcomps):

            aind = int(x.nodes.treenode_id[i])
            bind = int(x.nodes.parent_id[i])

            axyz = x.nodes[x.nodes.treenode_id == aind][["x", "y", "z"]].values
            bxyz = x.nodes[x.nodes.treenode_id == bind][["x", "y", "z"]].values

            complength[i] = np.sqrt(np.sum((axyz - bxyz) ** 2))

            meandiam = (
                x.nodes[x.nodes.treenode_id == aind].radius.values
                + x.nodes[x.nodes.treenode_id == bind].radius.values
            ) * 0.5

            area = (meandiam ** 2) / (4.0 * np.pi)

            M[tn_to_index[bind], tn_to_index[aind]] = (
                -area / complength[i] / Ra * 10 ** (-4)
            )
            M[tn_to_index[aind], tn_to_index[bind]] = M[
                tn_to_index[bind], tn_to_index[aind]
            ]

        complength[0] = complength[1]

        gleak = (compdiam * np.pi * complength) / (Rm * 10 ** 8)
        memcap = (compdiam * np.pi * complength) * Cm * (10 ** -6) / (10 ** 8)

        for i in range(nofcomps):
            M[i, i] = gleak[i] - np.sum(M[i])

        M = sparse.csr_matrix(M)

        if solve:

            M_solved = np.linalg.inv(sparse.csr_matrix.todense(M))
            return (M_solved, memcap)

        return (M, memcap)

    else:

        raise ValueError("Unknown object type!")


def current_injection(conducM: np.mat, curramp: float = 1e-11):

    """
    Calculates the conductance matrix for a given neuron

    Paramters
    ---------
    conducM:            numpy.matrix
                        The conductance matrix produced by

    curramp:            float
                        The amplitude of the injection current

    Returns
    --------

    Vm_mat:             numpy.matrix
                        A node-by-node matrix of relative membrane voltage (Vm)
                        given current injection

    Examples
    --------



    """

    nofcomps = conducM.shape[0]
    Vm_mat = np.zeros((nofcomps, nofcomps))

    for i in range(0, nofcomps):

        currinj = np.zeros(nofcomps)
        currinj[i] = curramp

        Vm = spsolve(conducM, currinj)

        Vm_mat[i, :] = Vm

    return Vm_mat


def dbs_func(x, eps: float = 1.0, min_samples: int = 10):

    """
    Parameters
    --------
    x:              A matrix to perform DBSCAN on

    eps:            float
                    The radius of the search circle

    min_samples:    int
                    The minimum number of samples to include in a cluster

    Returns
    --------
    labels:         cluster label of each datapoint
    n_clusters_:    number of unique clusters
    n_noise_:       number of noise points

    Examples
    --------

    """

    db = DBSCAN(eps=eps, min_samples=min_samples).fit(x)
    labels = db.labels_
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    n_noise_ = list(labels).count(-1)

    return (labels, n_clusters_, n_noise_)


def find_clusters(x, phate_operator, eps: float = 1.0, min_samples: int = 10):

    """
    Parameters
    --------
    x:              A matrix to perform DBSCAN on

    phate_operator: phate.operator object
                    The operator used to perform the dimensionality reduction
                    The number of dimensions to reduce to
                    is specified in this operator object

    eps:            float
                    Radius of the search circle

    min_samples:    int
                    Minimum number of samples to include in a cluster

    Returns
    --------
    mat_red:        Matrix wih dimensions reduced
    labels:         Cluster label of each datapoint
    n_clusters_:    Number of unique clusters
    n_noise_:       Number of noise points

    Examples
    --------

    """

    # performing PHATE dimensionality reduction
    mat_red = phate_operator.fit_transform(x)

    # performing DBSCAN
    labels, n_clusters_, n_noise_ = dbs_func(mat_red, eps, min_samples)

    return (mat_red, labels, n_clusters_, n_noise_)


def cluster_palette(cmap: str = "hsv", n_objects: int = 10, shuffle: bool = False):

    """
    Parameters
    --------
    cmap:           str
                    The colour pallete to be used

    n_objects:      int
                    number of ojects to create an individual colour for.
                    E.g. number of clusters

    shuffle:        bool
                    Whether to shuffle the colourmap
                    or to keep the colours of nearby objects the similar

    Returns
    --------
    label_to_col:   a dictionary where the keys are the label and the
                    values are the colour (rgb, 0-1)

    Examples
    --------

    """

    pal = sns.color_palette(cmap, n_objects)

    if shuffle:

        random.shuffle(pal)

    label_to_col = dict(zip([i for i in range(0, n_objects)], pal))
    label_to_col[-1] = (0.0, 0.0, 0.0)

    return label_to_col


def match_connectors_to_nodes(synapse_connections, neuron, synapse_type = 'post'):

    if synapse_type == 'post':

        u = synapse_connections[['x_post','y_post','z_post']].values
        v = neuron.postsynapses[['x','y','z']].values

        data = ssd.cdist(u, v, metric = 'euclidean')

        dist_mat = pd.DataFrame(index = [i for i in range(0, synapse_connections.shape[0])],
                                columns = [i for i in range(0, neuron.postsynapses.shape[0])],
                                data = data)



        ind = [np.argmin(dist_mat.iloc[i, :]) for i in range(0, synapse_connections.shape[0])]

        syn_con = synapse_connections.copy()
        syn_con['connector'] = [neuron.postsynapses.iloc[i, :].connector_id for i in ind]

        # creating a connector to node dictionary

        c2n = dict(zip(neuron.postsynapses.connector_id.tolist(), neuron.postsynapses.node_id.tolist()))
        syn_con['node'] = syn_con.connector.map(c2n)

    elif synapse_type == 'pre':

        u = synapse_connections[['x_pre','y_pre','z_pre']].values
        v = neuron.presynapses[['x','y','z']].values

        data = ssd.cdist(u, v, metric = 'euclidean')

        dist_mat = pd.DataFrame(index = [i for i in range(0, synapse_connections.shape[0])],
                                columns = [i for i in range(0, neuron.presynapses.shape[0])],
                                data = data)



        ind = [np.argmin(dist_mat.iloc[i, :]) for i in range(0, synapse_connections.shape[0])]

        syn_con = synapse_connections.copy()
        syn_con['connector'] = [neuron.presynapses.iloc[i, :].connector_id for i in ind]

        # creating connector to node dictionary

        c2n = dict(zip(neuron.presynapses.connector_id.tolist(), neuron.presynapses.node_id.tolist()))
        syn_con['node'] = syn_con.connector.map(c2n)


    return(syn_con)

def permute_start_end(test, test_in_roi, cluster):

    start_end = [i for i in itertools.permutations(test.nodes[test_in_roi][test.nodes[test_in_roi].node_cluster == cluster].node_id.tolist(), 2)]

    return(start_end)

def cluster_to_all_nodes(neuron, start_end_node_pairs):

    g = nbd.create_graph_structure(neuron, returned_object = 'graph')
    g_rev = g.reverse()

    nodes_of_cluster = []

    for i, j in start_end_node_pairs:

        paths = nx.all_simple_paths(g_rev, source = i, target = j)

        for path in paths:

            nodes_of_cluster.append(path)

    nodes_of_cluster = list(np.unique(list(chain.from_iterable(nodes_of_cluster))))
    return(nodes_of_cluster)


def quick_optimisation(x, min_samples_start):

    """
    x : a neuronlist
    min_samples_start: the minimum number of samples to start off with

    """
    min_samples_optim_values = []
    err = []

    for i in x:

        print(i.id, i.name)

        n = i.copy()

        try:

            n_prep = prepare_neuron(n, change_units = True, factor = 1e3)

        except IndexError:

            err.append(i.id)

            continue

        n_m, n_memcap = calculate_M_mat(n_prep, solve = False)
        n_m_solved = np.linalg.inv(sparse.csr_matrix.todense(n_m))

        phate_operator = phate.PHATE(n_components = 3, n_jobs = -2, verbose = False)

        min_sample_range = [i for i in range(0, min_samples_start + 1)][::-1]

        for j in min_sample_range:

            mat_red, labels, n_clust, n_noise = nbm.find_clusters(n_m_solved, phate_operator, eps=1e-02, min_samples = j)

            if n_noise == 0:

                min_samples_optim_values.append((i.id, j))

                break

            else:

                continue

    return(min_samples_optim_values, err)

def find_compartments_in_roi(neuron, roi, min_samples):

    n = neuron.copy()
    n_prep = prepare_neuron(n, change_units=True, factor = 1e3)
    n_m, n_memcap = calculate_M_mat(n_prep, solve = False)
    n_m_solved = np.linalg.inv(sparse.csr_matrix.todense(n_m))

    phate_operator = phate.PHATE(n_components=3, n_jobs=-2, verbose = False)

    mat_red, labels, n_clusters, n_noise = nbm.find_clusters(n_m_solved, phate_operator, eps = 1e-02, min_samples = min_samples)

    node_to_label = dict(zip(range(0, len(labels)), labels))

    n_prep.nodes['node_cluster'] = n_prep.nodes.index.map(node_to_label)

    n_in_roi = navis.in_volume(n_prep.nodes[['x','y','z']] * 1e3, roi, mode = 'IN', inplace = False)

    return(n_prep, n_in_roi)

def matching_inputs_to_compartments(neuron_id, roi):

    # Fetch the healed skeleton
    full_skeleton = nvneu.fetch_skeletons(neuron_id, heal = True)[0]
    # which of the whole neuron's nodes are in the roi
    skeleton_in_roi = navis.in_volume(full_skeleton.nodes[['x','y','z']], roi, mode='IN')

    # Fetch the neurons synapses
    postsyn = nvneu.fetch_synapse_connections(target_criteria = neuron_id)
    # match the connectors to nodes
    syn_to_node = match_connectors_to_nodes(postsyn, full_skeleton, synapse_type = 'post')
    # which of these are in the roi
    roi_syn_con = syn_to_node[syn_to_node.node.isin(full_skeleton.nodes[skeleton_in_roi].node_id.tolist())].copy()

    # Count how many synapses the upstream neurons have on your neuron of interest
    n_syn = dict(Counter(roi_syn_con.bodyId_pre.tolist()).most_common())

    # fetch these neurons
    a, b = nvneu.fetch_neurons(roi_syn_con.bodyId_pre.unique())
    a['n_syn'] = [n_syn[i] for i in a.bodyId.tolist()]

    # adding the instance names to the synapses in the roi
    bid_to_instance = dict(zip(a.bodyId.tolist(), a.instance.tolist()))
    roi_syn_con['instance'] = [bid_to_instance[i] for i in roi_syn_con.bodyId_pre]

    # compartmentalise the neuron and find the nodes in each compartment for the prepared neuron
    compartments_in_roi, nodes_in_roi = find_compartments_in_roi(full_skeleton, ca, 6)
    clusters = compartments_in_roi.nodes.node_cluster.unique()

    cluster_dict = {}

    # find the nodes that make up each compartment in the full neuron
    for i in clusters:

        clust_nodes = cluster_to_all_nodes(full_skeleton, start_end_node_pairs=permute_start_end(compartments_in_roi, nodes_in_roi, cluster = i))
        cluster_dict[i] = clust_nodes

    #cluster_dict = {k : s for s, k in cluster_dict.items()}
    #roi_syn_con['compartment'] = [cluster_dict[i] for i in roi_syn_con.node.tolist()]

    return(cluster_dict)
    #return(roi_syn_con)

    #return(full_skeleton, skeleton_in_roi, skeleton_prep, roi_syn_con, a)


def find_compartments_of_missing_nodes(roi_syn_con, nodes_with_cluster, full_neuron, ds_neuron):

    """
    DS neuron -- downsampled_neuron
    """

    # find the connections associated to nodes that do not have a cluster
    nodes_to_query = roi_syn_con[~roi_syn_con.node.isin(nodes_with_cluster)].node.tolist()

    # create a geodesic matrix between the nodes that do not have a compartment and the original, non-downsampled neuron
    gmat = navis.geodesic_matrix(full_neuron, nodes_to_query)

    # subset the above geodesic matrix columns to only contain the nodes in the downsampled neuron
    reduced_gmat = gmat.T[np.isin(gmat.T.index, ds_neuron.nodes.node_id.tolist())].T.copy()

    # find the node which is the smallest distance away from the unclassified node
    closest_classified_node = list(reduced_gmat.columns[np.argmin(reduced_gmat.values, axis = 1)])

    closest_classified_node_to_cluster = ds_neuron.nodes[ds_neuron.nodes.node_id.isin(np.unique(closest_classified_node))]
    closest_classified_node_to_cluster = dict(zip(closest_classified_node_to_cluster.node_id, closest_classified_node_to_cluster.node_cluster))

    query_nodes_to_closest_node = dict(zip(np.unique(nodes_to_query), np.array(reduced_gmat.columns[np.argmin(reduced_gmat.values, axis = 1)])))

    comps = [closest_classified_node_to_cluster[query_nodes_to_closest_node[i]] for i in query_nodes_to_closest_node.keys()]

    query_nodes_to_compartment = dict(zip(query_nodes_to_closest_node.keys(), comps))

    return(query_nodes_to_compartment)


def node_to_compartment_full_neuron(original_neuron, ds_neuron):

    """
    DS neuron - downsampled neuron (one that has been run through find_compartments_in_roi function)
    """

    # nodes to query
    ntq = np.isin(original_neuron.nodes.node_id.tolist(), ds_neuron.nodes.node_id.tolist())
    ntq = original_neuron.nodes[~ntq].node_id.tolist()

    # geodesic matrix
    gmat = navis.geodesic_matrix(original_neuron, ntq)

    #subset the columns to the nodes that are in the downsampled neuron
    red_gmat = gmat.T[np.isin(gmat.T.index, ds_neuron.nodes.node_id)].T.copy()

    ntc = dict(zip(red_gmat.index.tolist(),
                  list(chain.from_iterable([ds_neuron.nodes[ds_neuron.nodes.node_id == i].node_cluster.tolist() for i in red_gmat.columns[np.argmin(red_gmat.values, axis = 1)]]))))

    ntc = {**ntc, **dict(zip(ds_neuron.nodes.node_id.tolist(), ds_neuron.nodes.node_cluster.tolist()))}

    original_neuron.nodes['node_cluster'] = [ntc[i] for i in original_neuron.nodes.node_id]

    return(original_neuron)

def compartmentalise_neuron(neuron_id, roi):

    # Fetching the neuron
    ds_neuron = nvneu.fetch_skeletons(neuron_id, heal = True)[0]
    original_neuron = nvneu.fetch_skeletons(neuron_id, heal = True)[0]

    # Electrotonic model
    DS_NEURON = prepare_neuron(ds_neuron, change_units = True, factor = 1e3)
    test_m, test_memcap = calculate_M_mat(DS_NEURON, solve = False)
    test_m_solved = np.linalg.inv(sparse.csr_matrix.todense(test_m))

    # running PHATE
    phate_operator = phate.PHATE(n_components = 3, n_jobs = -2, verbose = False)

    mat_red, labels, n_clusters, n_noise = nbm.find_clusters(test_m_solved,
                                                              phate_operator,
                                                              eps = 1e-02,
                                                              min_samples = 6)

    # index and labels
    index_to_label = dict(zip(range(0, len(labels)), labels))
    ds_neuron.nodes['node_cluster'] = ds_neuron.nodes.index.map(index_to_label)

    node_to_compartment = dict(zip(ds_neuron.nodes.node_id.tolist(),
                                   ds_neuron.nodes.node_cluster.tolist()))

    unique_compartments = ds_neuron.nodes.node_cluster.unique()

    # Finding the cluster of the nodes that were removed when downsampling the neuron
    whole_neuron_node_to_cluster = []

    for i in unique_compartments:

        start_end = [i for i in itertools.permutations(ds_neuron.nodes[ds_neuron.nodes.node_cluster == i].node_id.tolist(), 2)]

        nodes_of_cluster = cluster_to_all_nodes(original_neuron, start_end)

        node_to_cluster_dictionary = dict(zip(nodes_of_cluster, [i] * len(nodes_of_cluster)))

        whole_neuron_node_to_cluster.append(node_to_cluster_dictionary)

    whole_neuron_node_to_cluster_dict = {k : v for d in whole_neuron_node_to_cluster for k, v in d.items()}

    #Fetching postsynapses
    ds_neuron_postsynapses = nvneu.fetch_synapse_connections(target_criteria=neuron_id)

    ds_neuron_synapse_to_node = match_connectors_to_nodes(ds_neuron_postsynapses,
                                                          original_neuron,
                                                          synapse_type = 'post')

    #Which nodes are in the CA?
    skeleton_in_roi = navis.in_volume(original_neuron.nodes[['x','y','z']].values, roi, inplace = False)
    roi_syn_con = ds_neuron_synapse_to_node[ds_neuron_synapse_to_node.node.isin(original_neuron.nodes[skeleton_in_roi].node_id.tolist())].copy()

    a, b = nvneu.fetch_neurons(roi_syn_con.bodyId_pre.unique())
    bid_to_instance = dict(zip(a.bodyId.tolist(), a.instance.tolist()))
    roi_syn_con['instance'] = [bid_to_instance[i] for i in roi_syn_con.bodyId_pre]

    nodes_to_query = roi_syn_con[~roi_syn_con.node.isin(whole_neuron_node_to_cluster_dict.keys())].node.tolist()
    comp_of_missing_nodes = find_compartments_of_missing_nodes(roi_syn_con,
                                                               list(whole_neuron_node_to_cluster_dict.keys()),
                                                               original_neuron,
                                                               ds_neuron)

    whole_neuron_node_to_cluster_dict = {**whole_neuron_node_to_cluster_dict, **comp_of_missing_nodes}

    roi_syn_con['compartment'] = [whole_neuron_node_to_cluster_dict[i] for i in roi_syn_con.node.tolist()]

    original_neuron = node_to_compartment_full_neuron(original_neuron, ds_neuron)

    return(original_neuron, ds_neuron, roi_syn_con)



from typing import Optional, Union, Any

import seaborn as sns
import random


def define_color_palette(cmap, n_clusters, shuffle = False):

    pal = sns.color_palette(cmap, n_clusters)

    if shuffle:

        random.shuffle(pal)

    label_to_col = dict(zip([i for i in range(0, n_clusters)], pal))
    label_to_col[-1] = (0.0, 0.0, 0.0)

    return(label_to_col)
